import subprocess
import os
import abc
import hashlib
import tarfile
import zipfile
import glob
from PIL import Image
from tqdm import tqdm
from torchvision import transforms, datasets
import numpy as np
import logging
from disentanglement_lib.data.unsupervised.unsupervied_data import UnsupervisedData, preprocess
from torchvision import transforms, datasets
import gin

ROOT=os.environ.get("DISENTANGLEMENT_LIB_DATA", ".")

class Chairs(datasets.ImageFolder, UnsupervisedData):
    """Chairs Dataset from [1].

    Notes
    -----
    - Link : https://www.di.ens.fr/willow/research/seeing3Dchairs

    Parameters
    ----------
    root : string
        Root directory of dataset.

    References
    ----------
    [1] Aubry, M., Maturana, D., Efros, A. A., Russell, B. C., & Sivic, J. (2014).
        Seeing 3d chairs: exemplar part-based 2d-3d alignment using a large dataset
        of cad models. In Proceedings of the IEEE conference on computer vision
        and pattern recognition (pp. 3762-3769).

    """

    urls = {"train": "https://www.di.ens.fr/willow/research/seeing3Dchairs/data/rendered_chairs.tar"}
    files = {"train": "chairs_64"}
    img_size = (1, 64, 64)

    def __init__(self,
                 logger=logging.getLogger(__name__)):
        self.train_data = os.path.join(self.root, 'chairs', type(self).files["train"])
        self.transforms = transforms.Compose([transforms.Grayscale(),
                                              transforms.ToTensor()])
        self.logger = logger

        if not os.path.isdir(self.train_data):
            self.logger.info("Downloading {} ...".format(str(type(self))))
            self.download()
            self.logger.info("Finished Downloading.")

        datasets.ImageFolder.__init__(self, self.train_data, transform=self.transforms)
        UnsupervisedData.__init__(self)

    def download(self):
        """Download the dataset."""
        save_path = os.path.join(self.root, 'chairs.tar')
        os.makedirs(self.train_data)
        subprocess.check_call(["curl", type(self).urls["train"],
                               "--output", save_path])

        self.logger.info("Extracting Chairs ...")
        tar = tarfile.open(save_path)
        tar.extractall(self.root)
        tar.close()
        os.rename(os.path.join(self.root, 'rendered_chairs'), self.train_data)

        os.remove(save_path)

        self.logger.info("Preprocessing Chairs ...")
        preprocess(os.path.join(self.train_data, '*/*'),  # root/*/*/*.png structure
                   size=type(self).img_size[1:],
                   center_crop=(400, 400))